/*
 * Copyright (C) 1996 Julian Byrne
 * Permission to reproduce for non-commercial purposes granted.
 *
 * Title:        Spam Wrapper
 * File name:    gspam.c
 * Version:      1.1
 * Usage:        setenv QUERY_STRING "a=?&c=?&h=?"; gspam
 * Description:  Provides a wrapper between a form on gspam.html
 *               and some network utility programs.
 * Dependencies: CGI compliant web server
 * Author:       Julian Byrne
 * Internet:     julian.byrne@eng.monash.edu.au (NO JUNK MAIL!)
 * Web:          http://kryten.eng.monash.edu.au/~jcb
 * History:      96/ 7/26 Initial version
 *               96/ 8/16 Cleaned up and new options added.
 * Notes:
 *   See Progs[] and Ops[] for a list of network facilities accessed.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>

#define NUL '\0'
#define false (0!=0)
#define true  (0==0)
#define MAXSTR (80)

#define loop for (;;) /* Endless loop */

typedef int bool;

int    ARGC; /* Make main() arguments globally accessable */
char **ARGV;
char **ENVP;

char *env_var = "QUERY_STRING";
char *back_msg=
  "\n\nClick on your browser's <Back> button to return to previous page.\n\n";

bool do_header = true; /* Still need to send HTML header */
char *p0, *p; /* Pointer to start & middle of QUERY_STRING env var */

char domain[MAXSTR];
char host[MAXSTR];
char athost[1+MAXSTR] = "@";

/* Programs list */

char *prog0[] = {"/usr/local/bin/dig", domain, "mx", NULL};
char *prog1[] = {"/usr/local/bin/dig", domain, "a", NULL};
char *prog2[] = {"/usr/local/bin/dig", "-x", domain, NULL};
char *prog3[] = {"/usr/local/bin/dig", domain, "hinfo", NULL};
char *prog4[] = {"/usr/etc/traceroute",
                 "-m", "40", "-v", "-w", "10", domain, NULL};
char *prog5[] = {"/usr/etc/ping", "-c", "10", domain, NULL};
char *prog6[] = {"/usr/bsd/finger", domain, NULL};
char *prog7[] = {"/usr/local/bin/whois", "-h", host, domain, NULL};
char *prog8[] = {"/usr/local/bin/dig", athost, domain, "axfr", NULL};

char **Progs[] = {prog0,prog1,prog2,prog3,prog4,prog5,prog6,prog7,prog8};

/* Legal arguments list */

char finger_legal[] =
"@ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_0123456789.";
char domain_legal[] =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_0123456789.";
char ip_legal[]=                                       "0123456789.";

/* Operations list */
char *Ops[][2]={
  /*  0 */ {"DIG hostname->mailexchanger",domain_legal},
  /*  1 */ {"DIG hostname->ipaddress",domain_legal},
  /*  2 */ {"DIG ipaddress->hostname",ip_legal},
  /*  3 */ {"DIG hostname hardware info", domain_legal},
  /*  4 */ {"TRACEROUTE hostname",domain_legal},
  /*  5 */ {"PING hostname",domain_legal},
  /*  6 */ {"FINGER user@hostname",finger_legal},
  /*  7 */ {"WHOIS international",finger_legal},
  /*  8 */ {"DIG domain list",domain_legal},
  /*  9 */ {"WHOIS US domain", "http://rs.internic.net/cgi-bin/whois?%s\n\n"},
  /* 10 */ {"WHOIS Europe domain", "http://www.ripe.net/cgi-bin/whois?%s\n\n"},
  /* 11 */ {"Pages in domain",   "http://altavista.digital.com/\
cgi-bin/query?pg=q&what=web&fmt=.&q=%%2Bhost%%3A%s\n\n"},
  /* 12 */ {"Links to domain",   "http://altavista.digital.com/\
cgi-bin/query?pg=q&what=web&fmt=.&q=%%2Blink%%3A%s+-host%%3A%s\n\n"},
  /* 13 */ {"Any ref's in web",  "http://altavista.digital.com/\
cgi-bin/query?pg=q&what=web&fmt=&q=%s\n\n"},
  /* 14 */ {"Any ref's in news", "http://altavista.digital.com/\
cgi-bin/query?pg=q&what=news&fmt=&q=%s\n\n"},
  /* 15 */ {"http://domain/",    "http://%s/\n\n"},
  /* 16 */ {"http://www.domain/","http://www.%s/\n\n"},
  /* 17 */ {"ftp://domain/",     "ftp://%s/\n\n"},
  /* 18 */ {"ftp://ftp.domain/", "ftp://ftp.%s/\n\n"},
  /* 19 */ {"telnet://domain/",      "telnet://%s/\n\n"},
  /* 20 */ {"telnet://domain:smtp/", "telnet://%s:25/\n\n"},
  /* 21 */ {"telnet://domain:nntp/", "telnet://%s:119/\n\n"},
  /* 22 */ {"abuse@domain",     "mailto:abuse@%s\n\n"},
  /* 23 */ {"admin@domain",     "mailto:admin@%s\n\n"},
  /* 24 */ {"root@domain",      "mailto:root@%s\n\n"},
  /* 25 */ {"postmaster@domain","mailto:postmaster@%s\n\n"},
  /* 26 */ {"info@domain",      "mailto:info@%s\n\n"},
  /* 27 */ {"support@domain",   "mailto:support@%s\n\n"},
  /* 28 */ {NULL,NULL}};

/* Convert ASCII hex digit to binary */

int xtod(char x)
{
  return (x >= 'a' && x <= 'z') ? (x-'a'+10) :
         (x >= 'A' && x <= 'Z') ? (x-'A'+10) :
         (x >= '0' && x <= '9') ? (x-'0'+ 0) : 0;
}

/* Convert 2 ASCII hex digits to binary */

int x2tod(char x, char y)
{
  return xtod(x)*16+xtod(y);
}

/* Send default HTML header if not already sent */

void header(void)
{
  if (do_header) {
    do_header = false;
    printf("Content-Type: text/plain\nPragma: no-cache\n\n"); }
}

/* Display error message with bad character pointer (if available)
   and then exit */

void fatal(char *format, ...)
{
  va_list args;

  header();

  va_start(args, format);
  vprintf(format, args);
  va_end(args);

  if (p != NULL)
    printf("\n\n  \"%s\"\n%*c", p0, p-p0+4, '^');

  printf("%s", back_msg);

  fflush(stdout);
  fflush(stderr);

  exit(EXIT_FAILURE);
}

/* Get next string token pointed to by p, place new p position in q,
   place token in tok making sure it's no longer than n.
   A token is delimited by &, = and end-of-string and has leading
   and trailing spaces but not embedded spaces stripped.
   Handles HTML character quoting conventions. */

void token(char *p, char **q, char *tok, int n)
{
  int i, j;

  while (isspace(*p)) /* Strip leading spaces */
    p++;

  for (j = i = 0; *p != '=' && *p != '&' && *p != NUL && i < n-1; i++, p++)
    if (*p == '%')
      if (p[1] == '%') {
        tok[i] = '%';
        p++; }
      else
      if (isxdigit(p[1]) && isxdigit(p[2])) {
        tok[i] = x2tod(p[1],p[2]);
        p     += 2; }
      else
        fatal("Error: Bad %% hex escape.");
    else
    if (*p == '+')
      tok[j = i] = ' ';
    else
      tok[i] = *p;

  while (i > j+1 && isspace(tok[i-1])) /* Strip trailing spaces */
    i--;

  tok[i] = NUL;
  *q     = p;
}

/* Converts a string s to t ready for HTML transport by
   quoting any special character. t must be three times
   the size of s to avoid possible memory corruption */

void armor(char *s, char *t)
{
  static char dtox[16]="0123456789ABCDEF";

  for (; *s != NUL; s++, t++)
    if (*s == ' ')
      *t = '+';
    else
    if (strchr("&;/#?: =+%", *s)) {
      *t++ = '%';
      *t++ = dtox[(*s)/16];
      *t   = dtox[(*s)%16]; }
    else
      *t = *s;

  *t = NUL;
}

/* Strip username from a domain. Naive users keep thinking an
   email address is a domain... */

char *strip_username(char *s)
{
  char *p = strrchr(s, '@');

  return (p == NULL) ? s : (p+1);
}

/* Parse environment variable QUERY_STRING and perform
   requested operation, sending results in HTML form to
   standard output. */

int main(int argc, char *argv[], char *envp[])
{
  int i, j, k;
  char command[MAXSTR], address[MAXSTR], v[MAXSTR], w[MAXSTR], x[MAXSTR*3];

  ARGC = argc;
  ARGV = argv;
  ENVP = envp;

  if (argc != 1)
    fatal("Error: Uses %s environment variable, not arguments.", env_var);

  p0 = p = getenv(env_var);

  if (p == NULL)
    fatal("Error: Contact server administrator, no %s environment variable.",
      env_var);

/* Defaults */

  strcpy(address, "");
  strcpy(command, Ops[0][0]);
  strcpy(host, "whois.internic.net");

/* Parse QUERY_STRING */

  loop {
    token(p, &p, v, sizeof(v));

    if (*p == '=') {
      p++;
      token(p, &p, w, sizeof(w)); }
    else
      w[0] = NUL;

    if (!strcasecmp(v, "a"))
      strcpy(address, w); /* Overwrites any previous value */
    else
    if (!strcasecmp(v, "c"))
      strcpy(command, w); /* Overwrites any previous value */
    else
    if (!strcasecmp(v, "h"))
      strcpy(host, w); /* Overwrites any previous value */
    /* Ignoring unrecognized variables */

    if (*p == NUL)
      break;
    if (*p != '&')
      fatal("Error: & or end-of-string expected.");

    p++; }

  strcpy(athost+1, host);

  p = NULL;

  j = strlen(command);
  for (i = 0; Ops[i][0] != NULL && (j < (k = strlen(Ops[i][0])) ||
    strcasecmp(command+j-k, Ops[i][0])); i++)
    ;

  if (Ops[i][0] == NULL) {
    armor(command, x);
    fatal("Error: Unrecognized command \"c=%s\".\n\
Are you using a cached, out-of-date copy of the web form?\n\
Try clicking your web browser's <Reload> button after clicking <Back>.", x); }

  /* Strip username@ for all except `finger' and international `WHOIS' */

  strcpy(domain, (i == 6 || i == 7) ? address : strip_username(address));

  if (i < sizeof(Progs)/sizeof(char **)) {

    if (domain[0] == NUL)
      fatal(
"Error: You must enter a host name, domain name or numerical IP address\n\
       before submitting a query.");

    j = 0;
    if (domain[0] == '-' ||
      (j = strspn(domain, Ops[i][1])) != strlen(domain))
      fatal(
  "Error: Illegal character in host/domain/IP address:\n\n  \"%s\"\n%*c",
      domain, j+4, '^');

    if (host[0] == NUL)
      fatal("Error: You must enter a host name before submitting a query.");

    j = 0;
    if (host[0] == '-' || (j = strspn(host, domain_legal)) != strlen(host))
      fatal("Error: Illegal character in host name:\n\n  \"%s\"\n%*c",
        host, j+4, '^');

    header();
    printf("\n%s results. %s", Ops[i][0], back_msg);
    fflush(stdout);
    fflush(stderr);

    execv(Progs[i][0], Progs[i]);

    fatal("Error: Accessing \"%s\". Please contact server administrator.",
      Progs[i][0]); }
  else {
    armor(domain, x);
    printf("Location: ");
    printf(Ops[i][1], x, x);
    do_header = false; }

  return errno;
}
